﻿//==========================================================
//	linksim-util.js by LinkSim for Web
//==========================================================
//#####################################################################
//	座標変換等
//#####################################################################
//***********************************************************
//	角度変換
//***********************************************************
const DegToRad = (deg) => deg * ( Math.PI / 180 );
const RadToDeg = (rad) => rad * ( 180 / Math.PI );

//***********************************************************
//	回転後のＸ座標
//		x0, y0 : 回転中心
//		xS, yS : 元の座標
//		a      : 回転角
//***********************************************************
function rotateX(x0, y0, xS, yS, a) {
  return x0+(xS-x0)*Math.cos(DegToRad(a)) - (yS-y0)*Math.sin(DegToRad(a));
}

//***********************************************************
//回転後のＹ座標
//  x0, y0 : 回転中心
//  xS, yS : 元の座標
//  a      : 回転角
//***********************************************************
function rotateY(x0, y0, xS, yS, a) {
  return  y0+(xS-x0)*Math.sin(DegToRad(a)) + (yS-y0)*Math.cos(DegToRad(a));
}

//***********************************************************
//  直近のグリッド上の座標
//***********************************************************
function FitGrid(val) {
	return Math.round(val / 10) * 10;
}

//***********************************************************
//  ２円の交点の座標を求める
//***********************************************************
function GetCircleToCircleCrossPoint(x1, y1, r1, x2, y2, r2) {
	//------------------------------------------------------------------
	function Revo(x1, y1, xc, yc, a) {
		this.x2 = xc + (x1-xc) * Math.cos(a) - (y1-yc) * Math.sin(a);
		this.y2 = yc + (x1-xc) * Math.sin(a) + (y1-yc) * Math.cos(a);
	}
	//------------------------------------------------------------------

	var ang;
	var bx1, by1, d;
	var a, h;
	var ax1, ay1, ax2, ay2;
	var rLong;

	ang = Math.atan2((y2 - y1), (x2 - x1));		// ２中心点間の傾き角度[rad]

	res = new Revo(x2, y2, x1, y1, -ang);		// 円２中心点を逆回転
	bx1 = res.x2;
	by1 = res.y2;

	d = Math.abs(bx1 - x1);						// ２中心点間の距離

	if (d > (r1 + r2) || d < Math.abs(r1 - r2) || d == 0) {
		if (r1 > r2) {
			rLong = r1;
		} else {
			rLong = r2;
		}

		if (d > rLong && d <= (r1*1.001 + r2*1.001)) {
			r1 = d - r2;
		} else if (d < rLong && d >= Math.abs(r1*1.001 - r2)) {
			r2 = r1 + d;
		} else if (d < rLong && d >= Math.abs(r1 - r2*1.001)) {
			r1 = r2 + d;
		} else {
			this.x1 = 0;
			this.y1 = 0;
			this.x2 = 0;
			this.y2 = 0;
			this.ResultCode = 0;				//解なし（交わらない）
			return;
		}
	}

	if (d == (r1 + r2) || d == Math.abs(r1 - r2)) {
		a = r1;
		if (r1 < r2) {
			a = d - r2;
		}
		res = new Revo((x1 + a), y1, x1, y1, ang);		// 回転
		ax1 = res.x2;
		ay1 = res.y2;
		this.x1 = ax1;
		this.y1 = ay1;
		this.x2 = ax1;
		this.y2 = ay1;
		this.ResultCode = 1           //重解（接する）
		
	} else {
		a = (d * d + r1 * r1 - r2 * r2) / 2.0 / d;
		h = Math.sqrt(r1 * r1 - a * a);
		res = new Revo((x1 + a), (y1 + h), x1, y1, ang);   // 回転
		ax1 = res.x2;
		ay1 = res.y2;
		
		res = new Revo((x1 + a), (y1 - h), x1, y1, ang);   // 回転
		ax2 = res.x2;
		ay2 = res.y2;

		this.x1 = ax1;
		this.y1 = ay1;
		this.x2 = ax2;
		this.y2 = ay2;
		this.ResultCode = 2;          //２箇所で交わる
	}
	
	return;
}

//***********************************************************
//  円と直線の交点の座標を求める
//***********************************************************
function GetCircleToLineCrossPoint(x1, y1, r1, x2, y2, r2) { //: TMathResult;
	var bx2, by2, d;
	var a, h;
	var ax1, ay1, ax2, ay2;
	var rLong;

	bx2 = rotateX(x1, y1, x2, y2, -r2);
	by2 = rotateY(x1, y1, x2, y2, -r2);

	d = Math.abs(by2 - y1);					//円と直線の距離

	if (d > r1) {
		if (d <= r1*1.001) {
			r1 = d;
		} else {
			this.x1 = 0;
			this.y1 = 0;
			this.x2 = 0;
			this.y2 = 0;
			this.ResultCode = 0;			//解なし（交わらない）
			return
		}
	}

	if (d == r1) {
		ax1 = rotateX(x1, y1, x1, by2, r2);
		ay1 = rotateY(x1, y1, x1, by2, r2);

		this.x1 = ax1;
		this.y1 = ay1;
		this.x2 = ax1;
		this.y2 = ay1;
		this.ResultCode = 1;				//重解（接する）
	} else {
		d = Math.sqrt(r1**2 - (by2 - y1)**2);
		ax1 = rotateX(x1, y1,  d+x1, by2, r2);
		ay1 = rotateY(x1, y1,  d+x1, by2, r2);
		ax2 = rotateX(x1, y1, -d+x1, by2, r2);
		ay2 = rotateY(x1, y1, -d+x1, by2, r2);

		this.x1 = ax1;
		this.y1 = ay1;
		this.x2 = ax2;
		this.y2 = ay2;
		this.ResultCode = 2;				//２箇所で交わる
	}
}

//***********************************************************
//  部品（parts index）の角度補正（0～359.99...度の間にする）
//***********************************************************
function AdjustAngle(idx) {
	while (parts[idx].a >= 360) {
		parts[idx].a = parts[idx].a - 360;
	}
	while (parts[idx].a < 0) {
		parts[idx].a = parts[idx].a + 360;
	}
}

//***********************************************************
//  部品（angle）の角度補正（0～359.99...度の間にする）
//***********************************************************
function AdjustAngleForAngle(angle) {
	while (angle >= 360) {
		angle = angle - 360;
	}
	while (angle < 0) {
		angle = angle + 360;
	}
	return angle;
}


//#####################################################################
//	図面の状態取得
//#####################################################################
//***********************************************************
//  ２点が同一の場所か否か？
//***********************************************************
function CmpPosition(x1, y1, x2, y2) {
	if (Math.abs(x1 - x2) <= 0.5 && Math.abs(y1 - y2) <= 0.5) {
		return true;
	} else {
		return false;
	}
}

function CmpPosition2(x1, y1, x2, y2, r) {
	if (Math.abs(x1 - x2) <= r && Math.abs(y1 - y2) <= r) {
		return true;
	} else {
		return false;
	}
}

//***********************************************************
//  現在位置に既にピンがあるか否か？　あった場合はピンタイプを返す
//***********************************************************
function IsPin(xs, ys) {
	var idx;
	var i;
	var x, y, angle;
	var xD, yD;
	var r, p;

	result = -1;     //ピンは無い

	for (idx = 0; idx < partsCount; idx++) {
		x = parts[idx].x;
		y = parts[idx].y;
		angle = parts[idx].a;
		r = parts[idx].r;

		//91～94 = piston
		if (parts[idx].kind == "piston1Power" || parts[idx].kind == "piston2Power" || parts[idx].kind == "piston1" || parts[idx].kind == "piston2") {
			while (r > 360) {
				r = r - 360;
			}
			while (r < 0) {
				r = r + 360;
			}
			if (r > 180) {
				p = 360 - r;
			} else {
				p = r;
			}
		} else {
		    p = 0;
		}

		for (i = 0; i < parts[idx].pinCount; i++) {
			xD = rotateX(x, y, x+parts[idx].pinX[i]+p, y+parts[idx].pinY[i], angle);
			yD = rotateY(x, y, x+parts[idx].pinX[i]+p, y+parts[idx].pinY[i], angle);

			if (CmpPosition(xD, yD, xs, ys) == true) {
				result = parts[idx].pinType[i];
				return result;
			}
		}
	}
	return result;
}

//***********************************************************
//  現在位置に既にペンがあるか否か？　あった場合はペンタイプを返す
//***********************************************************
function IsPen(xs, ys) {
	var idx;
	var i;
	var x, y, angle;
	var xD, yD;
	var r, p;

	result = "";     //ペンは無い

	for (idx = 0; idx < partsCount; idx++) {
		x = parts[idx].x;
		y = parts[idx].y;
		angle = parts[idx].a;
		r = parts[idx].r;

		//91～94 = piston
		if (parts[idx].kind == "piston1Power" || parts[idx].kind == "piston2Power" || parts[idx].kind == "piston1" || parts[idx].kind == "piston2") {
			while (r > 360) {
				r = r - 360;
			}
			while (r < 0) {
				r = r + 360;
			}
			if (r > 180) {
				p = 360 - r;
			} else {
				p = r;
			}
		} else {
		    p = 0;
		}

		for (i = 0; i < parts[idx].penCount; i++) {
			xD = rotateX(x, y, x+parts[idx].penX[i]+p, y+parts[idx].penY[i], angle);
			yD = rotateY(x, y, x+parts[idx].penX[i]+p, y+parts[idx].penY[i], angle);

			if (CmpPosition(xD, yD, xs, ys) == true) {
				result = parts[idx].penType[i];
				return result;
			}
        }
    }
	return result;
}

//***********************************************************
//  指定されたパーツ内の、近く（offsetドット以内）の連結穴の座標を取得
//***********************************************************
function GetNearPinPosition(idx, xs, ys, offset) {		//TPointD;
	var i;
	var x, y, r, angle;
	var xD, yD;
	var p;

	this.x = xs;
	this.y = ys;

	x = parts[idx].x;
	y = parts[idx].y;
	r = parts[idx].r;
	angle = parts[idx].a;

	//11, 12, 13, 21, 22, 23:
	if (parts[idx].kind == "disk1Power" || parts[idx].kind == "disk2Power" || parts[idx].kind == "disk3Power" ||
		parts[idx].kind == "disk1" || parts[idx].kind == "disk2" || parts[idx].kind == "disk3") {

		if (Math.abs(x - xs) <= offset && Math.abs(y - ys) <= offset) {
			this.x = x;
			this.y = y;
			return;
		}

		for (i = 0; i <= Math.round((r-40) / 20); i++) {
			xD = rotateX(x, y, x, y-(i*20+30), angle);
			yD = rotateY(x, y, x, y-(i*20+30), angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				return;
			}

			xD = rotateX(x, y, x, y+(i*20+30), angle);
			yD = rotateY(x, y, x, y+(i*20+30), angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				return;
			}

			xD = rotateX(x, y, x-(i*20+20), y, angle);
			yD = rotateY(x, y, x-(i*20+20), y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				return;
			}

			xD = rotateX(x, y, x+(i*20+20), y, angle);
			yD = rotateY(x, y, x+(i*20+20), y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				return;
			}
		}
	}

	//15, 16:
	if (parts[idx].kind == "rod1Power" || parts[idx].kind == "rod2Power") {
		for (i = 2; i <= Math.round((r-40) / 20); i++) {
			xD = rotateX(x, y, x + i*20, y, angle);
			yD = rotateY(x, y, x + i*20, y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				return;
			}
		}
	}

	//91, 92:
	if (parts[idx].kind == "piston2Power" || parts[idx].kind == "piston2") {
		while (r > 360) {
			r = r - 360;
		}
		while (r < 0) {
			r = r + 360;
		}
		if (r > 180) {
			p = 360 - r;
		} else {
			p = r;
		}
		for (i = 0; i <= 3; i++) {
			xD = rotateX(x, y, x + i*20 + 240 + p, y, angle);
			yD = rotateY(x, y, x + i*20 + 240 + p, y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				return;
			}
		}
	}

	//93, 94:
	if (parts[idx].kind == "piston1Power" || parts[idx].kind == "piston1") {
		while (r > 360) {
			r = r - 360;
		}
		while (r < 0) {
			r = r + 360;
		}
		if (r > 180) {
			p = 360 - r;
		} else {
			p = r;
		}
		for (i = 0; i <= 3; i++) {
			xD = rotateX(x, y, x + i*20 + 60 + p, y, angle);
			yD = rotateY(x, y, x + i*20 + 60 + p, y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				return;
			}
		}
	}

	//31, 32, 33, 34, 35, 36:
	if (parts[idx].kind == "rod1" || parts[idx].kind == "rod2" || parts[idx].kind == "rod3" || parts[idx].kind == "rod4" || parts[idx].kind == "rod5" || parts[idx].kind == "rod6") {
		for (i = 0; i <= Math.round((r-40) / 20); i++) {
			xD = rotateX(x, y, x + i*20, y, angle);
			yD = rotateY(x, y, x + i*20, y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				return;
			}
		}
	}

	//95, 96:
	if (parts[idx].kind == "rodSlide1" || parts[idx].kind == "rodSlide2") {
		if (parts[idx].kind == "rodSlide2") {
			r = 440;
		} else {
			r = 280;
		}
		for (i = 0; i < Math.round((r-40) / 20); i++) {
			if (i < 3 || i >= Math.round((r-40) / 20) - 3) {
				xD = rotateX(x, y, x + i*20, y, angle);
				yD = rotateY(x, y, x + i*20, y, angle);
				if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
					this.x = xD;
					this.y = yD;
					return;
				}
			}
		}
	}

	//41, 42, 43:
	if (parts[idx].kind == "triangle1" || parts[idx].kind == "triangle2") {
        for (i = 0; i < Math.round((r-40) / 20); i++) {
			xD = rotateX(x, y, x, y-(i*20), angle);
			yD = rotateY(x, y, x, y-(i*20), angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				return;
			}
		}

		for (i = 1; i < Math.round((r-40) / 20); i++) {
			xD = rotateX(x, y, x+(i*20), y, angle);
			yD = rotateY(x, y, x+(i*20), y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				return;
			}
		}

        for (i = 0; i <= Math.round((r-40) / 20); i++) {
			xD = rotateX(x, y, x+(i*20), y-(Math.round((r-40) / 20)-i)*20, angle);
			yD = rotateY(x, y, x+(i*20), y-(Math.round((r-40) / 20)-i)*20, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				return;
			}
		}
	}
}

//***********************************************************
//  現在位置の連結穴のカウント（パーツ数）
//***********************************************************
function CountHole(xs, ys) {
	var idx;
	var i;
	var x, y, r, angle;
	var xD, yD;
	var cnt;
	var p;

	cnt = 0;

	for (idx = 0; idx < partsCount; idx++) {

		x = parts[idx].x;
		y = parts[idx].y;
		r = parts[idx].r;
		angle = parts[idx].a;

		//円盤(11, 12, 13, 21, 22, 23)
		if (parts[idx].kind == "disk1Power" || parts[idx].kind == "disk2Power" || parts[idx].kind == "disk3Power" || parts[idx].kind == "disk1" || parts[idx].kind == "disk2" || parts[idx].kind == "disk3") {
			for (i = 0; i <= Math.round((r-40) / 20); i++) {
				xD = rotateX(x, y, x, y-(i*20+30), angle);
				yD = rotateY(x, y, x, y-(i*20+30), angle);
				if (CmpPosition2(xD, yD, xs, ys, 0.05) == true) cnt++;

				xD = rotateX(x, y, x, y+(i*20+30), angle);
				yD = rotateY(x, y, x, y+(i*20+30), angle);
				if (CmpPosition2(xD, yD, xs, ys, 0.05) == true) cnt++;

				xD = rotateX(x, y, x-(i*20+20), y, angle);
				yD = rotateY(x, y, x-(i*20+20), y, angle);
				if (CmpPosition2(xD, yD, xs, ys, 0.05) == true) cnt++;

				xD = rotateX(x, y, x+(i*20+20), y, angle);
				yD = rotateY(x, y, x+(i*20+20), y, angle);
				if (CmpPosition2(xD, yD, xs, ys, 0.05) == true) cnt++;
			}
		}
		
		//動力ロッド(15, 16)
		else if (parts[idx].kind == "rod1Power" || parts[idx].kind == "rod2Power") {
			for (i = 2; i <= Math.round((r-40) / 20); i++) {
				xD = rotateX(x, y, x + i*20, y, angle);
				yD = rotateY(x, y, x + i*20, y, angle);
				if (CmpPosition2(xD, yD, xs, ys, 0.05) == true) cnt++;
			}
		}

		//長いピストン(91, 92)
		else if (parts[idx].kind == "piston2Power" || parts[idx].kind == "piston2") {
			while (r > 360) {
				r = r - 360;
			}
			while (r < 0) {
				r = r + 360;
			}
			if (r > 180) {
				p = 360 - r;
			} else {
				p = r;
			}
			for (i = 0; i <= 3; i++) {
				xD = rotateX(x, y, x + i*20 + 240 + p, y, angle);
				yD = rotateY(x, y, x + i*20 + 240 + p, y, angle);
				if (CmpPosition2(xD, yD, xs, ys, 0.05) == true) cnt++;
			}
		}

		//短いピストン(93, 94)
		else if (parts[idx].kind == "piston1Power" || parts[idx].kind == "piston1") {
			while (r > 360) {
				r = r - 360;
			}
			while (r < 0) {
				r = r + 360;
			}
			if (r > 180) {
				p = 360 - r;
			} else {
				p = r;
			}

			for (i = 0; i <= 3; i++) {
				xD = rotateX(x, y, x + i*20 + 60 + p, y, angle);
				yD = rotateY(x, y, x + i*20 + 60 + p, y, angle);
				if (CmpPosition2(xD, yD, xs, ys, 0.05) == true) cnt++;
			}
		}

		//ロッド(31, 32, 33, 34)
		else if (parts[idx].kind == "rod1" || parts[idx].kind == "rod2" || parts[idx].kind == "rod3" || parts[idx].kind == "rod4" || parts[idx].kind == "rod5" || parts[idx].kind == "rod6") {
			for (i = 0; i <= Math.round((r-40) / 20); i++) {
				xD = rotateX(x, y, x + i*20, y, angle);
				yD = rotateY(x, y, x + i*20, y, angle);
				if (CmpPosition2(xD, yD, xs, ys, 0.05) == true) cnt++;
			}
		}

		//スライドロッド(95, 96)
		else if (parts[idx].kind == "rodSlide1" || parts[idx].kind == "rodSlide2") {
			if (parts[idx].kind == "rodSlide2") {
				r = 440;
			} else {
				r = 280;
			}
			
			for (i = 0; i <= Math.round((r-40) / 20); i++) {
				if (i < 3 || i > Math.round((r-40) / 20) - 3) {
					xD = rotateX(x, y, x + i*20, y, angle);
					yD = rotateY(x, y, x + i*20, y, angle);
					if (CmpPosition2(xD, yD, xs, ys, 0.05) == true) cnt++;
				}
			}
		}

		//三角(41, 42)
		else if (parts[idx].kind == "triangle1" || parts[idx].kind == "triangle2") {
			for (i = 0; i < Math.round((r-40) / 20); i++) {
				xD = rotateX(x, y, x, y-(i*20), angle);
				yD = rotateY(x, y, x, y-(i*20), angle);
				if (CmpPosition2(xD, yD, xs, ys, 0.05) == true) cnt++;
			}
			for (i = 1; i < Math.round((r-40) / 20); i++) {
				xD = rotateX(x, y, x+(i*20), y, angle);
				yD = rotateY(x, y, x+(i*20), y, angle);
				if (CmpPosition2(xD, yD, xs, ys, 0.05) == true) cnt++;
			}
			for (i = 0; i <= Math.round((r-40) / 20); i++) {
				xD = rotateX(x, y, x+(i*20), y-(Math.round((r-40) / 20)-i)*20, angle);
				yD = rotateY(x, y, x+(i*20), y-(Math.round((r-40) / 20)-i)*20, angle);
				if (CmpPosition2(xD, yD, xs, ys, 0.05) == true) cnt++;
			}
		}
	}

	return cnt;
}

//***********************************************************
//  指定されたパーツ内の、近く（offsetドット以内）の連結穴の座標を取得
//***********************************************************
function GetNearHolePosition(idx, xs, ys, offset) {
	var i;
	var x, y, r, angle;
	var xD, yD;
	var p;

	if (idx >= partsCount) {
		this.x = xs;
		this.y = ys;
		this.isHole = false;
		return;
	}

	x = parts[idx].x;
	y = parts[idx].y;
	r = parts[idx].r;
	angle = parts[idx].a;

	//円盤(11, 12, 13, 21, 22, 23)
	if (parts[idx].kind == "disk1Power" || parts[idx].kind == "disk2Power" || parts[idx].kind == "disk3Power" || parts[idx].kind == "disk1" || parts[idx].kind == "disk2" || parts[idx].kind == "disk3") {
		if (Math.abs(x - xs) <= offset && Math.abs(y - ys) <= offset) {
			this.x = x;
			this.y = y;
			this.isHole = true;
			return;
		}

		for (i = 0; i <= Math.round((r-40) / 20); i++) {
			xD = rotateX(x, y, x, y-(i*20+30), angle);
			yD = rotateY(x, y, x, y-(i*20+30), angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				this.isHole = true;
				return;
			}

			xD = rotateX(x, y, x, y+(i*20+30), angle);
			yD = rotateY(x, y, x, y+(i*20+30), angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				this.isHole = true;
				return;
			}

			xD = rotateX(x, y, x-(i*20+20), y, angle);
			yD = rotateY(x, y, x-(i*20+20), y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				this.isHole = true;
				return;
			}

			xD = rotateX(x, y, x+(i*20+20), y, angle);
			yD = rotateY(x, y, x+(i*20+20), y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				this.isHole = true;
				return;
			}
		}
	}

	//動力ロッド(15, 16)
	else if (parts[idx].kind == "rod1Power" || parts[idx].kind == "rod2Power") {
		for (i = 2; i <= Math.round((r-40) / 20); i++) {
			xD = rotateX(x, y, x + i*20, y, angle);
			yD = rotateY(x, y, x + i*20, y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				this.isHole = true;
				return;
			}
		}
	}

	//長いピストン(91, 92)
	else if (parts[idx].kind == "piston2Power" || parts[idx].kind == "piston2") {
		while (r > 360) {
			r = r - 360;
		}
		while (r < 0) {
			r = r + 360;
		}
		if (r > 180) {
			p = 360 - r;
		} else {
			p = r;
		}

		for (i = 0; i <= 3; i++) {
			xD = rotateX(x, y, x + i*20 + 240 + p, y, angle);
			yD = rotateY(x, y, x + i*20 + 240 + p, y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				this.isHole = true;
				return;
			}
		}
	}

	//短いピストン(93, 94)
	else if (parts[idx].kind == "piston1Power" || parts[idx].kind == "piston1") {
		while (r > 360) {
			r = r - 360;
		}
		while (r < 0) {
			r = r + 360;
		}
		if (r > 180) {
			p = 360 - r;
		} else {
			p = r;
		}

		for (i = 0; i <= 3; i++) {
			xD = rotateX(x, y, x + i*20 + 60 + p, y, angle);
			yD = rotateY(x, y, x + i*20 + 60 + p, y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				this.isHole = true;
				return;
			}
		}
	}

	//ロッド(31, 32, 33, 34)
	else if (parts[idx].kind == "rod1" || parts[idx].kind == "rod2" || parts[idx].kind == "rod3" || parts[idx].kind == "rod4" || parts[idx].kind == "rod5" || parts[idx].kind == "rod6") {
		for (i = 0; i <= Math.round((r-40) / 20); i++) {
			xD = rotateX(x, y, x + i*20, y, angle);
			yD = rotateY(x, y, x + i*20, y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				this.isHole = true;
				return;
			}
		}
	}

	//スライドロッド(95, 96)
	else if (parts[idx].kind == "rodSlide1" || parts[idx].kind == "rodSlide2") {
		if (parts[idx].kind == "rodSlide2") {
			r = 440;
		} else {
			r = 280;
		}
		
		for (i = 0; i <= Math.round((r-40) / 20); i++) {
			if (i < 3 || i > Math.round((r-40) / 20) - 3) {
				xD = rotateX(x, y, x + i*20, y, angle);
				yD = rotateY(x, y, x + i*20, y, angle);
				if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
					this.x = xD;
					this.y = yD;
					this.isHole = true;
					return;
				}
			}
		}
	}

	//三角(41, 42)
	else if (parts[idx].kind == "triangle1" || parts[idx].kind == "triangle2") {
		for (i = 0; i < Math.round((r-40) / 20); i++) {
			xD = rotateX(x, y, x, y-(i*20), angle);
			yD = rotateY(x, y, x, y-(i*20), angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				this.isHole = true;
				return;
			}
		}

        for (i = 1; i < Math.round((r-40) / 20); i++) {
			xD = rotateX(x, y, x+(i*20), y, angle);
			yD = rotateY(x, y, x+(i*20), y, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				this.isHole = true;
				return;
			}
		}

        for (i = 0; i <= Math.round((r-40) / 20); i++) {
			xD = rotateX(x, y, x+(i*20), y-(Math.round((r-40) / 20)-i)*20, angle);
			yD = rotateY(x, y, x+(i*20), y-(Math.round((r-40) / 20)-i)*20, angle);
			if (Math.abs(xD - xs) <= offset && Math.abs(yD - ys) <= offset) {
				this.x = xD;
				this.y = yD;
				this.isHole = true;
				return;
			}
		}
	}

	this.x = xs;
	this.y = ys;
	this.isHole = false;
	return;
}

//***********************************************************
//直近（5px以内）の接続穴の座標とParts[]Indexを取得（無い場合は、Parts[]Index=-0が返る）
//***********************************************************
function GetNearHoleByAllParts(x, y) {

	for (var idx = 0; idx < partsCount; idx++) {
		var pos = new GetNearHolePosition(idx, x, y, 5);
		
		//接続穴が見つかった
		if (pos.isHole == true) {
			this.x = pos.x;
			this.y = pos.y;
			this.idx = idx;
			return;
		}
	}
	
	//接続穴が見つからない
	this.x = x;
	this.y = y;
	this.idx = -1;
	return;
}


//***********************************************************
//  現在位置のピンに接続されている部品
//    Job = -1   部品数の取得
//    Job > -1   部品番号の取得（Job番目の部品番号）
//***********************************************************
function GetPartsNumberForPinPosition(xs, ys, job) {
	var idx;
	var i;
	var x, y, angle;
	var xD, yD;
	var cnt;
	var r, p;

	cnt = 0;     //部品は無い

	for (idx = 0; idx < partsCount; idx++) {
		x = parts[idx].x;
		y = parts[idx].y;
		angle = parts[idx].a;
		r = parts[idx].r;

        if (parts[idx].kind == "piston1Power" || parts[idx].kind == "piston2Power" || parts[idx].kind == "piston1" || parts[idx].kind == "piston2") {
			while (r > 360) {
				r = r - 360;
			}
			while (r < 0) {
				r = r + 360;
			}
			if (r > 180) {
				p = 360 - r;
			} else {
				p = r;
			}
		} else {
			p = 0;
		}

		for (i = 0; i < parts[idx].pinCount; i++) {
			xD = rotateX(x, y, x+parts[idx].pinX[i]+p, y+parts[idx].pinY[i], angle);
			yD = rotateY(x, y, x+parts[idx].pinX[i]+p, y+parts[idx].pinY[i], angle);

			if (CmpPosition(xD, yD, xs, ys) == true) {
				if (job == cnt) {
					return idx;
				}
				cnt++;
			}
		}
	}

	return cnt;
}

//***********************************************************
//  ピンインデックスに接続されている部品
//    部品番号no以外の部品の番号を返す		1 112 2
//		PartsCount: 読み飛ばしパーツ数
//***********************************************************
function GetPartsNumberForPinIndex(PartsNo, PinIndex, PartsCount) {

	var r = 0;
	var result = -1;

	if (PartsCount < 1) {
		for (var idx = 0; idx < partsCount; idx++) {
		
			if (idx == PartsNo) continue;

			for (var i = 0; i < parts[idx].pinCount; i++) {
				if (parts[idx].pinIndex[i] == PinIndex) {
					r++;
					//result = idx;
				}
			}
		}
		return r;//esult;

	} else {
		for (var idx = 0; idx < partsCount; idx++) {
			if (idx == PartsNo) continue;

			for (var i = 0; i < parts[idx].pinCount; i++) {

				if (parts[idx].pinIndex[i] == PinIndex) {
					result = idx;
					r++;
					if (r == PartsCount) {
						return result;
					}
				}
			}
		}

		return result;
	}
}

//***********************************************************
//  ピンポジションの取得
//***********************************************************
function GetPinPositionForPinNumber(idx, Pno) {
	var x, y, angle;
	var r, p;

	x = parts[idx].x;
	y = parts[idx].y;
	angle = parts[idx].a;
	r = parts[idx].r;

    if (parts[idx].kind == "piston1Power" || parts[idx].kind == "piston2Power" || parts[idx].kind == "piston1" || parts[idx].kind == "piston2") {
		while (r > 360) {
			r = r - 360;
		}
		while (r < 0) {
			r = r + 360;
		}
		if (r > 180) {
			p = 360 - r;
		} else {
			p = r;
		}
	} else {
		p = 0;
	}

	this.x = rotateX(x, y, x+parts[idx].pinX[Pno]+p, y+parts[idx].pinY[Pno], angle);
	this.y = rotateY(x, y, x+parts[idx].pinX[Pno]+p, y+parts[idx].pinY[Pno], angle);
}

//***********************************************************
//  ピンインデックスからピンポジションの取得
//***********************************************************
function GetPinPositionForPinIndex(PinIndex, PinType) {
	var p;

	for (var no = 0; no < partsCount; no++) {
		for (var pno = 0; pno < parts[no].pinCount; pno++) {
			if (parts[no].pinIndex[pno] == PinIndex && parts[no].pinType[pno] == PinType) {
				var x = parts[no].x;
				var y = parts[no].y;
				var angle = parts[no].a;
				var r = parts[no].r;

				if (parts[no].kind == "piston1Power" || parts[no].kind == "piston2Power" || parts[no].kind == "piston1" || parts[no].kind == "piston2") {
					while (r > 360) {
						r = r - 360;
					}
					while (r < 0) {
						r = r + 360;
					}
					if (r > 180) {
						p = 360 - r;
					} else {
						p = r;
					}
				} else {
					p = 0;
				}

				this.x = rotateX(x, y, x+parts[no].pinX[pno]+p, y+parts[no].pinY[pno], angle);
				this.y = rotateY(x, y, x+parts[no].pinX[pno]+p, y+parts[no].pinY[pno], angle);
				return;
			}
		}
	}
	
	this.x = -999999;
	this.y = -999999;
	return;
}

//***********************************************************
//  現在位置に既にピンがあるか否か？　あった場合はピン番号を返す
//***********************************************************
function GetPinNumber(idx, xs, ys) {
	var i;
	var x, y, angle;
	var xD, yD;
	var r, p;

	result = -1;     //ピンは無い

	x = parts[idx].x;
	y = parts[idx].y;
	angle = parts[idx].a;
	r = parts[idx].r;

    if (parts[idx].kind == "piston1Power" || parts[idx].kind == "piston2Power" || parts[idx].kind == "piston1" || parts[idx].kind == "piston2") {
		while (r > 360) {
			r = r - 360;
		}
		while (r < 0) {
			r = r + 360;
		}
		if (r > 180) {
			p = 360 - r;
		} else {
			p = r;
		}
	} else {
		p = 0;
	}

	for (i = 0; i < parts[idx].pinCount; i++) {
		xD = rotateX(x, y, x+parts[idx].pinX[i]+p, y+parts[idx].pinY[i], angle);
		yD = rotateY(x, y, x+parts[idx].pinX[i]+p, y+parts[idx].pinY[i], angle);

		if (CmpPosition(xD, yD, xs, ys) == true) {
			result = i;
			break;
		}
	}
	
	return result;
}
//---------------------------------------------------------------
//	partsT版
//---------------------------------------------------------------
function GetPinNumberT(idx, xs, ys) {
	var i;
	var x, y, angle;
	var xD, yD;
	var r, p;

	result = -1;     //ピンは無い

	x = partsT[idx].x;
	y = partsT[idx].y;
	angle = partsT[idx].a;
	r = partsT[idx].r;

    if (partsT[idx].kind == "piston1Power" || partsT[idx].kind == "piston2Power" || partsT[idx].kind == "piston1" || partsT[idx].kind == "piston2") {
		while (r > 360) {
			r = r - 360;
		}
		while (r < 0) {
			r = r + 360;
		}
		if (r > 180) {
			p = 360 - r;
		} else {
			p = r;
		}
	} else {
		p = 0;
	}

	for (i = 0; i < partsT[idx].pinCount; i++) {
		xD = rotateX(x, y, x+partsT[idx].pinX[i]+p, y+partsT[idx].pinY[i], angle);
		yD = rotateY(x, y, x+partsT[idx].pinX[i]+p, y+partsT[idx].pinY[i], angle);

		if (CmpPosition(xD, yD, xs, ys) == true) {
			result = i;
			break;
		}
	}
	
	return result;
}

//***********************************************************
//  指定した部品のピン番号をピンインデックスから得る
//***********************************************************
function GetPinNumberForPinIndex(PartsNo, PinIndex) {
	var result = -1;
	for (var i = 0; i < parts[PartsNo].pinCount; i++) {
		if (parts[PartsNo].pinIndex[i] == PinIndex) {
			result = i;
			break;
		}
	}
	return result;
}

//***********************************************************
//  自由度を求める
//    すべてのピンについて、その属性を変更することなく
//    検索を終了した場合のみ、Trueを返す。
//    いずれかのピンの属性を変更した場合は、
//    Falseを返すことで、呼出ルーチンが再度自由度の検査を行う。
//***********************************************************
function SetFreeLevel() {
	var i, j;
	var ii, jj;
	var FixPin1;
	var FixPin2;
	var Pos1, Pos2;				//: TPointD;

	result = true;

	//すべての部品についての処理（固定ピンにする処理）
	for (i = 0; i < partsCount; i++) {

		if (parts[i].kind == "disk1Power" || parts[i].kind == "disk2Power" || parts[i].kind == "disk3Power" ||
			parts[i].kind == "disk1" || parts[i].kind == "disk2" || parts[i].kind == "disk3" ||
			parts[i].kind == "rod1Power" || parts[i].kind == "rod2Power" ||
			parts[i].kind == "rod1" || parts[i].kind == "rod2" || parts[i].kind == "rod3" || parts[i].kind == "rod4" || parts[i].kind == "rod5" || parts[i].kind == "rod6" ||
			parts[i].kind == "rodSlide1" || parts[i].kind == "rodSlide2" ||
			parts[i].kind == "triangle1" || parts[i].kind == "triangle2") {

			FixPin1 = 0;
			FixPin2 = 0;

			//固定ピンの数の算出
			for (j = 0; j < parts[i].pinCount; j++) {
				if (parts[i].pinType[j] == 1) FixPin1++;
				if (parts[i].pinType[j] == 2) FixPin2++;
			}

			//固定ピンが２箇所以上ある場合は、動かせないので、
			//  この部品のすべてのピンを固定ピンにする。
			//  接続されている部品のこの位置のピンも固定属性とする。
			if (FixPin1 >= 2) {

				//すべてのピンについての処理
				for (j = 0; j < parts[i].pinCount; j++) {

					//可動ピンの場合は、この位置のピンをすべて固定ピンに
					if (parts[i].pinType[j] == 0) {

						Pos1 = new GetPinPositionForPinNumber(i, j);

						//現在のピン位置と同一の位置のピンをすべて固定ピンにする
						for (ii = 0; ii < partsCount; ii++) {

							for (jj = 0; jj < parts[i].pinCount; jj++) {

								if (parts[i].pinIndex[j] == parts[ii].pinIndex[jj]) {
									parts[ii].pinType[jj] = 2;
								}

								Pos2 = new GetPinPositionForPinNumber(ii, jj);

								if (CmpPosition2(Pos1.x, Pos1.y, Pos2.x, Pos2.y, 0) == true) {
									parts[ii].pinType[jj] = 1;
								}
							}
						}

						parts[i].pinType[j] = 1;
						result = false;
					}
				}

				parts[i].freeLevel = 0;						//移動不可能
				//固定ピンが２個以上のパーツは移動済みとする
				parts[i].moveFlag = 1;

			} else if (FixPin1 == 1 && FixPin2 > 0) {
				parts[i].freeLevel = 0;							//移動不可能

			} else if (FixPin1 + FixPin2 == 0) {
				parts[i].freeLevel = 2;							//移動不確定

			} else {
				parts[i].freeLevel = 1;							//移動可能
			}
		} else {
			parts[i].freeLevel = 1;								//移動可能
		}
	}
	
	return result;
}


//***********************************************************
//	ペン関連
//***********************************************************
//-----------------------------------------------------------
//  ペン位置リストをクリア
//-----------------------------------------------------------
function ClearPenPosition() {
	penBlackListX = [];
	penBlackListY = [];
	penBlueListX = [];
	penBlueListY = [];
	penRedListX = [];
	penRedListY = [];
	penGreenListX = [];
	penGreenListY = [];
}

//-----------------------------------------------------------
//  ペン位置をリストに取得
//-----------------------------------------------------------
function GetPenPosition() {
	var　idx, pno;
	var　x, y, angle;
	var　xD, yD;
	var　r, p;

	var pBlack = false;
	var pBlue = false;
	var pRed = false;
	var pGreen = false;

	for (idx = 0; idx < partsCount; idx++) {

		x = parts[idx].x;
		y = parts[idx].y;
		angle = parts[idx].a;
		r = parts[idx].r;

		if (parts[idx].kind == "piston1Power" || parts[idx].kind == "piston2Power" || 
			parts[idx].kind == "piston1" || parts[idx].kind == "piston2") {
			while (r > 360) {
				r = r - 360;
			}
			while (r < 0) {
				r = r + 360;
			}
			if (r > 180) {
				p = 360 - r;
			} else {
				p = r;
			}
		} else {
			p = 0;
		}

		for (pno = 0; pno < parts[idx].penCount; pno++) {
			xD = rotateX(x, y, x+parts[idx].penX[pno]+p, y+parts[idx].penY[pno], angle);
			yD = rotateY(x, y, x+parts[idx].penX[pno]+p, y+parts[idx].penY[pno], angle);


			if (parts[idx].penType[pno] == "black" && pBlack == false) {
				if (penBlackListX.length >= Math.floor(360 / sp)) {
					penBlackListX.shift();		//先頭要素の削除
					penBlackListY.shift();		//先頭要素の削除
				}
				penBlackListX.push(xD);
				penBlackListY.push(yD);
				pBlack = true;
			
			} else if (parts[idx].penType[pno] == "blue" && pBlue == false) {
				if (penBlueListX.length >= Math.floor(360 / sp)) {
					penBlueListX.shift();		//先頭要素の削除
					penBlueListY.shift();		//先頭要素の削除
				}
				penBlueListX.push(xD);
				penBlueListY.push(yD);
				pBlue = true;
			
			} else if (parts[idx].penType[pno] == "red" && pRed == false) {
				if (penRedListX.length >= Math.floor(360 / sp)) {
					penRedListX.shift();		//先頭要素の削除
					penRedListY.shift();		//先頭要素の削除
				}
				penRedListX.push(xD);
				penRedListY.push(yD);
				pRed = true;
			
			} else if (parts[idx].penType[pno] == "green" && pGreen == false) {
				if (penGreenListX.length >= Math.floor(360 / sp)) {
					penGreenListX.shift();		//先頭要素の削除
					penGreenListY.shift();		//先頭要素の削除
				}
				penGreenListX.push(xD);
				penGreenListY.push(yD);
				pGreen = true;
			}
		}
	}
}


//***********************************************************
//	パーツ配列の退避、復帰
//***********************************************************
//-------------------------------------------------------
//	parts[] →　partsB[]
//-------------------------------------------------------
function pushB() {
	for (var ii = 0; ii < partsCount; ii++) {
		partsB[ii] = Object.assign({}, parts[ii]);
		partsB[ii].pinType = Object.assign({}, parts[ii].pinType);
		partsB[ii].pinIndex = Object.assign({}, parts[ii].pinIndex);
		partsB[ii].pinX = Object.assign({}, parts[ii].pinX);
		partsB[ii].pinY = Object.assign({}, parts[ii].pinY);
		partsB[ii].penType = Object.assign({}, parts[ii].penType);
		partsB[ii].penX = Object.assign({}, parts[ii].penX);
		partsB[ii].penY = Object.assign({}, parts[ii].penY);
	}
}

//-------------------------------------------------------
//	parts[] →　partsT[]
//-------------------------------------------------------
function pushT() {
	for (var ii = 0; ii < partsCount; ii++) {
		partsT[ii] = Object.assign({}, parts[ii]);
		partsT[ii].pinType = Object.assign({}, parts[ii].pinType);
		partsT[ii].pinIndex = Object.assign({}, parts[ii].pinIndex);
		partsT[ii].pinX = Object.assign({}, parts[ii].pinX);
		partsT[ii].pinY = Object.assign({}, parts[ii].pinY);
		partsT[ii].penType = Object.assign({}, parts[ii].penType);
		partsT[ii].penX = Object.assign({}, parts[ii].penX);
		partsT[ii].penY = Object.assign({}, parts[ii].penY);
	}
}

//-------------------------------------------------------
//	parts[] →　partsArr[]
//-------------------------------------------------------
function pushPartsArr(partsArr) {
	for (var ii = 0; ii < partsCount; ii++) {
		partsArr[ii] = Object.assign({}, parts[ii]);
		partsArr[ii].pinType = Object.assign({}, parts[ii].pinType);
		partsArr[ii].pinIndex = Object.assign({}, parts[ii].pinIndex);
		partsArr[ii].pinX = Object.assign({}, parts[ii].pinX);
		partsArr[ii].pinY = Object.assign({}, parts[ii].pinY);
		partsArr[ii].penType = Object.assign({}, parts[ii].penType);
		partsArr[ii].penX = Object.assign({}, parts[ii].penX);
		partsArr[ii].penY = Object.assign({}, parts[ii].penY);
	}
}

//-------------------------------------------------------
//	partsB[] →　parts[]
//-------------------------------------------------------
function popB() {
	for (var ii = 0; ii < partsCount; ii++) {
		parts[ii] = Object.assign({}, partsB[ii]);
		parts[ii].pinType = Object.assign({}, partsB[ii].pinType);
		parts[ii].pinIndex = Object.assign({}, partsB[ii].pinIndex);
		parts[ii].pinX = Object.assign({}, partsB[ii].pinX);
		parts[ii].pinY = Object.assign({}, partsB[ii].pinY);
		parts[ii].penType = Object.assign({}, partsB[ii].penType);
		parts[ii].penX = Object.assign({}, partsB[ii].penX);
		parts[ii].penY = Object.assign({}, partsB[ii].penY);
	}
}

//-------------------------------------------------------
//	partsT[] →　parts[]
//-------------------------------------------------------
function popT() {
	for (var ii = 0; ii < partsCount; ii++) {
		parts[ii] = Object.assign({}, partsT[ii]);
		parts[ii].pinType = Object.assign({}, partsT[ii].pinType);
		parts[ii].pinIndex = Object.assign({}, partsT[ii].pinIndex);
		parts[ii].pinX = Object.assign({}, partsT[ii].pinX);
		parts[ii].pinY = Object.assign({}, partsT[ii].pinY);
		parts[ii].penType = Object.assign({}, partsT[ii].penType);
		parts[ii].penX = Object.assign({}, partsT[ii].penX);
		parts[ii].penY = Object.assign({}, partsT[ii].penY);
	}
}

//-------------------------------------------------------
//	partsArr[] →　parts[]
//-------------------------------------------------------
function popPartsArr(partsArr) {
	for (var ii = 0; ii < partsCount; ii++) {
		parts[ii] = Object.assign({}, partsArr[ii]);
		parts[ii].pinType = Object.assign({}, partsArr[ii].pinType);
		parts[ii].pinIndex = Object.assign({}, partsArr[ii].pinIndex);
		parts[ii].pinX = Object.assign({}, partsArr[ii].pinX);
		parts[ii].pinY = Object.assign({}, partsArr[ii].pinY);
		parts[ii].penType = Object.assign({}, partsArr[ii].penType);
		parts[ii].penX = Object.assign({}, partsArr[ii].penX);
		parts[ii].penY = Object.assign({}, partsArr[ii].penY);
	}
}


//===========================================================
//	デバッグ
//===========================================================
function myConsoleLog(data) {
	if (debugMode == true) {
		console.log(data);
	}
}

